# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END


__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


import json
import pandas as pd

from foglamp.plugins.filter.breakover.core.base import _LOGGER
from foglamp.plugins.filter.breakover.core.bucket_handler import (
    bucket_config_properties, bucket_config_default
)
from foglamp.plugins.filter.breakover.core.post_process_handler import PostProcess


PLUGIN_NAME = "breakover"
_grp1 = "Input"
_grp2 = "Inference"
_grp3 = "Post Process"
_grp4 = "Output"
_grp5 = "Enable"


DEFAULT_CONFIG = {
    "plugin": {
        "group": "Enable",
        "description": "Filter for breakover forecasting.",
        "type": "string",
        "default": PLUGIN_NAME,
        "readonly": "true",
    },
    "bucket_config": {
        "order": "1",
        "displayName": "ML Model",
        "description": "Breakover forecasting model to be imported from bucket.",
        "type": "bucket",
        "properties": bucket_config_properties,
        "default": json.dumps(bucket_config_default)
    },
    "inputs": {
        "order": "2",
        "group": _grp1,
        "displayName": "Inputs",
        "description": "Configure input data that is to be used for breakover forecasting.",
        "type": "list",
        "items": "object",
        "default": '[ {"asset": "asset_1", "datapoint": "datapoint_1", "feature_type": "Boolean"}]',
        "properties": {
            "asset": {
                "displayName": "Asset name",
                "description": "Enter input asset name.",
                "type": "string",
                "default": "asset_1",
            },
            "datapoint": {
                "displayName": "Datapoint name",
                "description": "Enter input datapoint name.",
                "type": "string",
                "default": "datapoint_1",
            },
            "feature_type": {
                "displayName": "Feature type",
                "description": "Datatype of the feature.",
                "type": "enumeration",
                "options": [
                    "Boolean",
                    "Numeric",
                ],
                "default": "Boolean",
            },
            "data_scaling": {
                "validity": 'feature_type == "Numeric"',
                "displayName": "Scaling strategy",
                "description": "The strategy that is to be used for data scaling.",
                "type": "enumeration",
                "options": [
                    "None",
                    "Standardization",
                    "Normalization",
                ],
                "default": "None",

            },
            "min_value": {
                "validity": 'feature_type == "Numeric" && data_scaling == "Normalization"',
                "displayName": "Minimum",
                "description": "Enter minimum value for data normalization.",
                "type": "float",
                "default": "-1.0",
            },
            "mean_value": {
                "validity": 'feature_type == "Numeric" && data_scaling == "Standardization"',
                "displayName": "Mean",
                "description": "Enter mean value for data standardization.",
                "type": "float",
                "default": "-1.0",
            },
            "max_value": {
                "validity": 'feature_type == "Numeric" && data_scaling == "Normalization"',
                "displayName": "Maximum",
                "description": "Enter maximum value for data normalization.",
                "type": "float",
                "default": "-1.0",
            },
            "std_value": {
                "validity": 'feature_type == "Numeric" && data_scaling == "Standardization"',
                "displayName": "Standard deviation",
                "description": "Enter standard deviation value for data standardization.",
                "type": "float",
                "default": "-1.0",
            },
        }
    },
    "sequence_length": {
        "order": "3",
        "group": _grp2,
        "displayName": "Sequence size",
        "description": "Sequence size that is to be used for breakover forecasting.",
        "type": "integer",
        "minimum": "10",
        "maximum": "300",
        "default": "10",

    },
    "inference_interval": {
        "order": "4",
        "group": _grp2,
        "displayName": "Inference interval",
        "description": "Time interval after which the data will be passed "
                        "through the model.",
        "type": "float",
        "default": "1.0",
    },
    "inference_interval_unit": {
        "order": "5",
        "group": _grp2,
        "displayName": "Inference interval unit",
        "description": "Time interval unit",
        "type": "enumeration",
        "options": [
            "seconds",
            "minutes",
            "hours",
        ],
        "default": "seconds",
    },

    "post_process_window_length": {
        "order": "6",
        "group": _grp3,
        "displayName": "Score Window Length",
        "description": "Number of score values that is to be used for breakover forecasting.",
        "type": "integer",
        "minimum": "10",
        "maximum": "300",
        "default": "10",
    },
    "post_process_measure": {
        "order": "7",
        "group": _grp3,
        "displayName": "Measure name",
        "description": "Measure to calculate on the scores value present in the window.",
        "type": "enumeration",
        "options": list(PostProcess.measures.keys()),
        "default": "Average",
    },
    "post_process_threshold": {
        "order": "8",
        "group": _grp3,
        "displayName": "Threshold",
        "description": "Threshold to apply on the calculated measure value.",
        "type": "float",
        "default": "0.5",
    },

    "forward_data": {
        "order": "9",
        "group": _grp4,
        "displayName": "Forward data",
        "description": "Whether to forward received PLC data down the filter chain.",
        "type": "boolean",
        "default": "true",
    },
    "output_asset_name": {
        "order": "10",
        "group": _grp4,
        "displayName": "Asset name",
        "description": "Name of asset that indicates breakover forecast.",
        "type": "string",
        "default": "breakover",
    },
    "output_score_datapoint_name": {
        "order": "11",
        "group": _grp4,
        "displayName": "Score datapoint name",
        "description": "Name of datapoint that indicates breakover forecast score.",
        "type": "string",
        "default": "forecast_score",
    },
    "output_forecast_datapoint_name": {
        "order": "12",
        "group": _grp4,
        "displayName": "Forecast datapoint name",
        "description": "Name of datapoint that indicates breakover forecast.",
        "type": "string",
        "default": "forecast_imminent",
    },
    "add_metadata": {
        "order": "13",
        "group": _grp5,
        "displayName": "Add performance counters",
        "description": "Add performance counters metadata containing resource utilization statistics.",
        "type": "boolean",
        "default": "false",
    },
    "log_level": {
        "order": "14",
        "group": _grp5,
        "displayName": "Log level",
        "description": "Log level for diagnostic or error output",
        "type": "enumeration",
        "options": ["ERROR", "WARN", "INFO", "DEBUG"],
        "default": "INFO",
    },
    "enable": {
        "order": "15",
        "group": _grp5,
        "displayName": "Enabled",
        "description": "Enable breakover forecast filter plugin",
        "type": "boolean",
        "default": "false",
    },

}

def _create_input_dict(info):
    """Cleanup the input config object.

    Args:
        info: single config object part of input list
    Returns:
        restructured_config: dict eg. below
            {
                "asset": "asset_1",            # mandatory
                "datapoint": "datapoint_1",    # mandatory
                "feature_type": "boolean",     # mandatory; options: `boolean`, `numeric`
                "scaling_strategy": None       # optional; options: None, `normalization`, `standardization`
                "minimum": 0.0,                # optional; with `scaling_strategy==normalization`
                "maximum": 100.0,              # optional; with `scaling_strategy==normalization`
                "mean": 50.0,                  # optional; with `scaling_strategy==standardization`
                "std": 0.5,                    # optional; with `scaling_strategy==standardization`
            }
    """
    feature_type = info["feature_type"].lower()
    scaler_strategy = (
        info["data_scaling"].lower() if "data_scaling" in info else "none"
    )

    supported_combination = {
        "boolean": ["none"],
        "numeric": ["none", "normalization", "standardization"],
    }
    if scaler_strategy not in supported_combination[feature_type]:
        _LOGGER.error(
            "The provided scaling strategy `{scaler_strategy}` isn't supported for `{feature_type}` feature type."
        )
        raise

    # Creating config
    d = {
        "asset": info["asset"],
        "datapoint": info["datapoint"],
        "feature_type": feature_type,
        "scaler_strategy": scaler_strategy if scaler_strategy != "none" else None
    }
    if d["scaler_strategy"] == "normalization":
        d["stats"] = {
            "minimum": float(info["min_value"]),
            "maximum": float(info["max_value"])
        }
    elif d["scaler_strategy"] == "standardization":
        d["stats"] = {
            "mean": float(info["mean_value"]),
            "std": float(info["std_value"])
        }
    return d


def parse_config(plugin_config):
    """Plugin configuration parser."""
    for key in ["inputs", "bucket_config"]:
        plugin_config[key]["value"] = json.loads(plugin_config[key]["value"])

    breakover_config = {
        "inputs": [
            _create_input_dict(info)
            for info in plugin_config["inputs"]["value"]
        ],
        "inference": {
            "sequence_length": int(plugin_config["sequence_length"]["value"]),
            "trigger_interval": pd.to_timedelta("{} {}".format(
                plugin_config["inference_interval"]["value"],
                plugin_config["inference_interval_unit"]["value"]
            )).total_seconds(),
            "bucket": {
                "logger": None,
                "bucket_client": None,
                "bucket_type": plugin_config["bucket_config"]["value"]["type"],
                "model_name": plugin_config["bucket_config"]["value"]["name"],
                "model_version": plugin_config["bucket_config"]["value"]["version"],
            }
        },
        "post_process": {
            "window_length": int(plugin_config["post_process_window_length"]["value"]),
            "measure": plugin_config["post_process_measure"]["value"],
            "threshold": float(plugin_config["post_process_threshold"]["value"]),
        },
        "output": {
            "asset": plugin_config["output_asset_name"]["value"],
            "score_datapoint": plugin_config["output_score_datapoint_name"]["value"],
            "forecast_datapoint": plugin_config["output_forecast_datapoint_name"]["value"],
            "forward_data": plugin_config["forward_data"]["value"] == "true",
        },
        "add_metadata": plugin_config["add_metadata"]["value"] == "true",
        "log_level": plugin_config["log_level"]["value"],
        "enable": plugin_config["enable"]["value"] == "true"
    }
    _LOGGER.debug(f"CUSTOM MESSAGE: `{str(type(breakover_config))}`  `{breakover_config}`")
    del plugin_config

    return breakover_config
