# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Plugin module to apply post processing to model score and finally convert to True/False forecast."""

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"

import gc
import os
import sys
import time
import numpy as np
# import pandas as pd
from typing import Union
from collections import deque

from foglamp.plugins.filter.breakover.core.base import _LOGGER
from foglamp.plugins.filter.breakover.core.base import DEFAULT_SCORE, DEFAULT_FORECAST


class PostProcess:
    """A class that apply post processing on the model score and finally convert to True/False."""

    __slots__ = ["logger", "window_size", "measure_name", "measure_func", "threshold", "_queue"]

    measures = {
        "Summation": np.sum,
        "Minimum": np.min,
        "Average": np.mean,
        "Median": np.median,
        "Maximum": np.max,

        "Standard Deviation": np.std,
        "Variance": np.var,
    }

    def __init__(
        self,
        window_size = 10,
        measure_name = "average",
        threshold = 0.5,
        logger = _LOGGER
    ):
        """Initialize Post Processing class.

        If -1 is observed in the data window then the prediction will be converted to -1.

        Args:
            window_size: size of the data window that hold the score.
            measure_name: Measure to calculate on the window of the data.
            threshold: Threshold value to convert the prediction to True/False.
            logger: logger instance to use

        """
        self.logger = _LOGGER
        self.window_size = window_size
        self._queue = deque(maxlen=self.window_size)
        self.measure_name = measure_name
        self.measure_func = self.measures[self.measure_name]
        self.threshold = threshold

    # @property
    # def allowed_measures(self):
    #     """Return the allowed measures."""
    #     return list(self.measures.keys())

    def register_score_value(self, curr_value: Union[int, float]):
        """Add value in the queue.

        Args:
            curr_value: Get the new reading for the series
        Return:

        """
        self._queue.append(curr_value)
        del curr_value

    def generate_forecast(self):
        """Calculate Simple Moving Measure.

        Return:
            the latest calculated value
        """
        forecast = DEFAULT_FORECAST
        if (
            (DEFAULT_SCORE in self._queue) or
            (len(self._queue) < self.window_size)
        ):
            moving_measure = None
        else:
            moving_measure = self.measure_func(self._queue)
            if moving_measure >= self.threshold:
                forecast = 1
            else:
                forecast = 0

        _msg = f"The calculated moving measure(`{self.measure_name}`) value "
        _msg += f"of breakover score is `{moving_measure}` which means, "
        if forecast == 0:
            _msg += "breakover is imminent."
        elif forecast == 1:
            _msg += "breakover is unlikely."
        else:
            _msg += "breakover can't be predicted."
        _msg += f" Size of score _queue: `{sys.getsizeof(self._queue)}`."
        self.logger.debug(_msg)

        del moving_measure, _msg
        gc.collect()

        return forecast
