# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END


__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


import gc
import time
from typing import Union

from foglamp.plugins.filter.breakover.core.base import (
    _LOGGER, LOG_LEVEL, TriggerExecution, TrackMemoryAllocation,
    DEFAULT_SCORE, DEFAULT_FORECAST
)
from foglamp.plugins.filter.breakover.config import parse_config
from foglamp.plugins.filter.breakover.core.bucket_handler import get_model_path
from foglamp.plugins.filter.breakover.core.data_handler import DataAccumulator
from foglamp.plugins.filter.breakover.core.inference_handler import Inference
from foglamp.plugins.filter.breakover.core.post_process_handler import PostProcess


class BreakoverExec:
    """Execute the  with FogLAMP."""

    __slots__ = [
        "add_metadata", "breakover_cfg", "data_manager", "enable", "logger",
        "predictor", "post_processor", "trigger", "last_poll_timestamp"
    ]

    def __init__(self, plugin_cfg, bucket_client):
        """Initialize breakover instance

        Args:
            plugin_cfg: foglamp configuration
            bucket_client: bucket client instance
        """
        self.breakover_cfg = parse_config(plugin_cfg)
        _LOGGER.info(f"breakover parsed config :  {self.breakover_cfg}")

        # re-adjust logging level & update config
        _LOGGER.setLevel(LOG_LEVEL[self.breakover_cfg["log_level"]])
        self.logger = _LOGGER

        self.breakover_cfg["inference"]["bucket"]["logger"] = self.logger
        self.breakover_cfg["inference"]["bucket"]["bucket_client"] = bucket_client

        self.add_metadata = self.breakover_cfg["add_metadata"]
        self.enable = self.breakover_cfg["enable"]
        self.logger.debug("Breakover Forecasting plugin is set to `Enable`.")

        if self.enable:
            self.initialize()

        self.last_poll_timestamp = None

    def initialize(self):
        """Initialize modules for enabling breakover forecasting."""
        # Initialize Data Preparation
        self.data_manager = DataAccumulator(
            self.breakover_cfg["inputs"],
            sequence_length=self.breakover_cfg["inference"]["sequence_length"],
            pad_sequence_length=5,
            logger=self.logger
        )

        # TODO - update the model initialization to handle scenario where bucket
        #        service might not have initialized at time of initialization.
        # get model path; updated logger is also shared
        model_path = get_model_path(**self.breakover_cfg["inference"]["bucket"])
        if model_path is None:
            # unable to get model from bucket service & error has been raise
            self.enable = False
            self.logger.info("BreakoverExec.initialize() Failed. Disabling the breakover plugin.")
            return

        # Initialize Inference; updated logger is also shared
        self.predictor = Inference(
            model_path = model_path,
            logger=self.logger
        )

        # Initialize Post Processor of Score
        self.post_processor = PostProcess(
            window_size = self.breakover_cfg["post_process"]["window_length"],
            measure_name = self.breakover_cfg["post_process"]["measure"],
            threshold = self.breakover_cfg["post_process"]["threshold"],
            logger = self.logger
        )

        # Initialize Trigger
        self.trigger = TriggerExecution(
            wait_period = self.breakover_cfg["inference"]["trigger_interval"],
            timer_refresh_interval = 0.01
        )

        self.logger.info("BreakoverExec.initialize() completed.")

    def get_logger(self):
        """Returns the updated logger instance."""
        return self.logger

    def run(self, data: Union[list, dict]):
        """ Execute breakover forecasting on new data.

        Args:
            data: readings data
        Return:
            data: updated data
        """
        if not self.enable:
            return

        poll_interval = (
            (time.time() - self.last_poll_timestamp)
            if self.last_poll_timestamp is not None else
            None
        )
        self.last_poll_timestamp = time.time()

        if isinstance(data, dict):
            data = [data]

        # add data for feature generation
        self.data_manager.filter_and_add_data(data)

        # calculate score if allowed & append to data
        reading = None
        if self.trigger.is_execution_allowed:

            self.trigger.reset_timer  # reset early, else inference time will get appended & delay will occur

            memory_tracker = TrackMemoryAllocation()
            with memory_tracker:

                data_array, max_user_ts, data_prep_time = None, -1.0, -1.0
                score, forecast, inference_time = DEFAULT_SCORE, DEFAULT_FORECAST, -1

                try:
                    data_array, max_user_ts, data_prep_time = self.data_manager.get_prepared_data()

                except Exception as e:
                    _msg = f"Error in breakover data preparation. `{str(e)}`. Skipping inference generation..."
                    self.logger.error(_msg)

                try:
                    if data_array is not None:
                        score, inference_time = self.predictor.predict(data_array)

                        # Adding reading for post processing & generating forecast
                        self.post_processor.register_score_value(score)
                        forecast = self.post_processor.generate_forecast()

                except Exception as e:
                    _msg = "Error in breakover forecasting. `{}`".format(str(e))
                    self.logger.error(_msg)

            memory_retained_at_end_MB, peak_memory_consumption_MB = memory_tracker.stats()

            _msg  = f"Prediction:: max_user_ts: {max_user_ts} "
            _msg += f"| poll interval   : {poll_interval:.3f} s "
            _msg += f"| memory retained : {memory_retained_at_end_MB:.3f} MB "
            _msg += f"| peak memory     : {peak_memory_consumption_MB:.3f} MB "
            _msg += f"| data prep time  : {data_prep_time:.3f} s "
            _msg += f"| inference time  : {inference_time:.3f} s "
            _msg += f"| score           : {score:.3f}\n"
            self.logger.debug(_msg)

            if data_array is not None:
                reading = {
                    "readings": {
                        self.breakover_cfg["output"]["score_datapoint"]: score,
                        self.breakover_cfg["output"]["forecast_datapoint"]: forecast
                    },
                    "asset": self.breakover_cfg["output"]["asset"],
                    "id": 0,
                    "ts": str(max_user_ts),
                    "user_ts": str(max_user_ts)
                }
                if self.add_metadata:
                    reading["readings"]["poll_interval_sec"] = poll_interval
                    reading["readings"]["data_preparation_time_sec"] = data_prep_time
                    reading["readings"]["inference_time_sec"] = inference_time
                    reading["readings"]["retained_memory_MB"] = memory_retained_at_end_MB
                    reading["readings"]["peak_memory_MB"] = peak_memory_consumption_MB

            del memory_tracker, memory_retained_at_end_MB, peak_memory_consumption_MB
            del data_array, max_user_ts, data_prep_time, score, forecast, inference_time

        # forward data
        if self.breakover_cfg["output"]["forward_data"]:
            if reading is not None:
                data.append(reading)
            else:
                # simply forward the data
                pass
        else:
            if isinstance(reading, dict):
                data = [reading]
            else:
                data = []

        _msg = f"A total of {len(data)} elements received in the data list."
        self.logger.debug(_msg)

        del reading, _msg
        gc.collect()

        return data

    def stop(self):
        """Stop the breakover forecasting instance."""
        self.trigger.shutdown_timer

        self.trigger = None
        self.breakover_cfg = None
        self.data_manager = None
        self.predictor = None
        self.post_processor = None
